from typing import Optional

import click

from grid.cli import rich_click
from grid.cli import observables
from grid.cli.client import Grid
from grid.cli.types import ObservableType
from grid.cli.utilities import is_experiment
from grid.sdk.rest.exceptions import GridException


def _check_is_experiment(ctx, _param, value):
    """Callback that checks if a value is an experiment."""
    if not value:
        return value

    if is_experiment(value):
        raise click.BadArgumentUsage(f"Must pass a Run name, not Experiment: {value}")

    return value


@rich_click.command()
@rich_click.argument('run', type=str, nargs=1, required=False, callback=_check_is_experiment)
@click.option(
    '--global', 'is_global', type=bool, is_flag=True, help='Fetch status from all collaborators when flag is passed'
)
def status(run: Optional[str] = None, export: Optional[str] = None, is_global: Optional[bool] = False) -> None:
    """Checks the status of Runs, Experiments, and Sessions."""
    # TODO - move the logic to this function and remove observables after session is fully migrated
    try:
        observables.Run().get(is_global=is_global, run_name=run)
    except GridException as e:
        raise click.ClickException(str(e))
    # If we have a Run, then don't print the global
    # interactive nodes table.
    if not run:
        client = Grid()
        client.status(kind=ObservableType.INTERACTIVE, identifiers=[], export=export, is_global=is_global)



import warnings
from datetime import datetime
import os
from grid.sdk.affirmations import affirm, is_not_deleted, is_not_created, is_not_shallow, is_created
from pathlib import Path
import textwrap
import time
from typing import List, Optional, Union
from urllib.parse import urlparse

import click

from grid.openapi import V1GetDatastoreResponse, V1DatastoreOptimizationStatus
from grid.sdk import env
from grid.sdk._gql.queries import delete_datastore, get_available_datastores, get_user_teams
from grid.sdk.client import create_swagger_client
from grid.sdk.rest import GridRestClient
from grid.sdk.rest.datastores import get_datastore_from_id, get_datastore_from_name
from grid.sdk.user import Team, User
from grid.sdk.utilities import MAGIC_NAME
from grid.sdk.utils.datastore_uploader import create_datastore_session

DATASTORE_TERMINAL_STATES = ['FAILED', 'SUCCEEDED']


def fetch_datastore(datastore_name: str, datastore_version: int, cluster: str) -> 'Datastore':
    """
    Validate the existence of provided datastore and user's access. Inject datastore id to the
    config, based on the name and version provided. If version is not provided, this function also
    injects the maximum version to the config
    # TODO - may be a better place for this logic
    """
    split = datastore_name.split(":")
    owner = None
    if len(split) > 2:
        raise ValueError(f"Error while parsing {datastore_name}. Use the format <username>:<datastore-name>")
    elif len(split) == 2:
        datastore_name = split[1]
        owner = split[0]

    if datastore_version is None:  # select the max version available
        possible_datastores = [d for d in list_datastores(is_global=True) if d.name == datastore_name]
        if len(possible_datastores) == 0:
            raise ValueError(
                f'No ready-to-use datastore found with name {datastore_name} '
                f'and version {datastore_version}'
            )
        selected_dstore = possible_datastores[0]
        for dstore in possible_datastores:
            if dstore.version > selected_dstore.version:
                selected_dstore = dstore
        warnings.warn(
            f'No ``--datastore_version`` passed. Using datastore: {datastore_name} version: {selected_dstore.version}'
        )
    else:
        selected_dstore = Datastore(name=datastore_name, version=datastore_version, owner=owner, cluster_id=cluster)

    return selected_dstore


class Datastore:
    _name: str
    _id: str
    _version: int
    _source: Optional[Union[str, Path]]
    _team: Optional[Team]
    _compression: bool
    _snapshot_status: str
    _created_at: datetime
    _user: User
    _team: Optional[Team]
    _cluster_id: str
    _size: str

    _is_deleted: bool
    _is_created: bool
    _is_shallow: bool

    def __init__(
        self,
        name: Optional[str] = None,
        source: Optional[os.PathLike] = None,
        team: Optional[Team] = None,
        user: Optional[User] = None,
        version: int = 1,
        compression: Optional[bool] = None,
        cluster_id: Optional[str] = None,
    ):
        """Initialize a new DataStore Object.

        If a DataStore with the given name, version, team and cluster already exists,
        then the object returned will be able to interact with the existing DataStore.

        Alternatively, if the DataStore is going to be created for the first time, then
        the ``source`` and ``compression`` parameters can be used to specify the location
        of the DataStore on disk (or at a remote location). and to optionally compress the
        data before uploading.

        After initializing the datastore object, the data itself can be uploaded by calling
        the ``upload()`` method.
        # TODO - user and team shouldn't be arguments
        Parameters
        ----------
        name
            The name of the DataStore.
        version
            The version of the DataStore.
        source
            The location of the DataStore on disk or at a remote location.
        team
            The name of the team that owns the DataStore.
        user
            The user that owns the DataStore.
        compression
            True if the DataStore should be compressed before uploading. Otherwise False.
        cluster_id
            The name of the cluster that the DataStore should be uploaded to.
        """

        # --------------------------------------------------------------------------- #
        #    This should be the first block that goes into the constructor of the     #
        #    resource object. This is to ensure that the resource object can be       #
        #    created from another internal call without filling the values and        #
        #    let __getattr__ picks up the instruction for refreshing values from ID   #
        self._client = GridRestClient(api_client=create_swagger_client())
        cluster_id = cluster_id or env.CONTEXT
        self._is_shallow = False
        self._cluster_id = cluster_id  # will be overwritten if comes from the config
        if name == MAGIC_NAME:
            self._is_shallow = True
            self._is_created = False
            self._is_deleted = False
            return
        #                                                                             #
        # --------------------------------------------------------------------------- #

        if name is None:
            if source:
                name = parse_name_from_source(source)
            else:
                raise ValueError("Name is required if source is not provided.")
        else:
            try:
                datastore: V1GetDatastoreResponse = get_datastore_from_name(
                    client=self._client, cluster_id=cluster_id, datastore_name=name, version=version
                )
                self._setup_from_response(datastore)
                return
            except KeyError:
                self._is_deleted = False  # the datastore has not been deleted
                self._is_created = False  # it doesn't exists in the grid backend.
                pass

        self._name = name
        self._version = version
        self._source = source
        self._team = team
        self._user = user
        self._compression = compression
        self._cluster_id = cluster_id
        self._id = None
        self._snapshot_status = None
        self._created_at = None
        self._size = None

    def _setup_from_response(self, datastore: V1GetDatastoreResponse):
        self._is_deleted = datastore.status.phase == V1DatastoreOptimizationStatus.DELETED
        self._is_created = True
        self._is_shallow = False

        self._id = datastore.id
        self._cluster_id = datastore.spec.cluster_id
        self._version = datastore.spec.version
        self._source = datastore.spec.source
        self._snapshot_status = str(datastore.status.phase)  # TODO - rename to status
        self._created_at = datastore.created_at
        self._size = f"{datastore.spec.size_mib} MiB"
        self._user = User(user_id=datastore.spec.user_id, username="", first_name="", last_name="")

    @classmethod
    def _from_existing(
        cls,
        name: str,
        version: int,
        owner: User,
        size: str,
        created_at: datetime,
        snapshot_status: str,
        datastore_id: str,
        cluster_id: str,
        team: Optional[Team] = None,
    ):
        # TODO - remove this and use _setup_from_response when we move from gQL to REST
        dstore = cls(name=MAGIC_NAME, cluster_id=cluster_id)
        dstore._name = name
        dstore._version = version
        dstore._id = datastore_id
        dstore._size = size
        dstore._created_at = created_at
        dstore._snapshot_status = snapshot_status
        dstore._user = owner
        dstore._team = team
        dstore._cluster_id = cluster_id

        dstore._source = None
        dstore._compression = None
        dstore._is_created = True
        dstore._is_shallow = False

        return dstore

    @classmethod
    def _from_id(cls, datastore_id: str, cluster_id: Optional[str] = env.CONTEXT) -> "Run":
        instance = cls(name=MAGIC_NAME, cluster_id=cluster_id)
        instance._id = datastore_id
        instance._is_shallow = True
        return instance

    @property
    def id(self) -> str:
        return self._id

    # ------------------ Attributes Only Valid Before Upload ---------------

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def source(self) -> Union[str, Path]:
        """The directory path at which the datastore is initialized from.

        !!! Note

            This property is only availabe to the instance of this class which uploads
            the datastore. Previously existing datastores will not posses any value
            for this property.
        """
        return self._source

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def compression(self) -> bool:
        """Boolean indicating if the data should be compressed during upload.

        If True, the uploaded data will be decompressed into it's original
        directory structure before the datastore is attached to a run or session.
        This is purely a measure to save time during upload (if the data is
        known to be compressible).

        !!! Note

            This property is only availabe to the instance of this class which uploads
            the datastore. Previously existing datastores will not posses any value
            for this property.

        """
        return self._compression

    # ------------------ Attributes Fully Active After Upload ---------------

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def name(self) -> str:
        """The name of the datastore.
        """
        return self._name

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def version(self) -> int:
        """The version of the datastore.
        """
        return self._version

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def user(self) -> User:
        """Information about the owner of the datastore (name, username, etc).
        """
        return self._user

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def team(self) -> Team:
        """Information about the team which owns the datastore

        !!! info

            This will only ever populate for Grid users who are enrolled in a
            "Teams" plan. Please see https://www.grid.ai/pricing/ for more info.
        """
        return self._team

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def created_at(self) -> datetime:
        """Date-Time timestamp when this datastore was created (first uploaded).
        """
        return self._created_at

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def size(self) -> str:
        """Size (in Bytes) of the datastore.
        """
        return self._size

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def snapshot_status(self) -> str:
        """The status of the datastore.
        """
        if (self._snapshot_status and
            (self._snapshot_status.upper() not in DATASTORE_TERMINAL_STATES)) or ((self._snapshot_status is None) and
                                                                                  (self._is_created is True)):
            self._update_status()
        return self._snapshot_status

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def cluster_id(self) -> str:
        """ID of the cluster which this datastore is uploaded to.

        !!! info

            This feature is only available to bring-your-own-cloud-credentials
            customers. Please see https://www.grid.ai/pricing/ for more info.
        """
        return self._cluster_id

    def _unshallow(self):
        if not self._is_shallow:
            raise RuntimeError('Datastore is already unshallow')
        if not hasattr(self, '_id') or self._id is None:
            raise RuntimeError("Cannot unshallow resource without a valid Datastore id")
        self._is_shallow = False
        run = get_datastore_from_id(self._client, datastore_id=self._id, cluster_id=self.cluster_id)
        self._setup_from_response(run)

    # -------------- Dunder Methods ----------------------

    @affirm(is_not_shallow, is_not_deleted)
    def __repr__(self):
        if self._is_created:
            res = textwrap.dedent(
                f"""\
                {self.__class__.__name__}(
                    {"name": <10} = \"{self.name}\",
                    {"version": <10} = {self.version},
                    {"size": <10} = \"{self.size}\",
                    {"created_at": <10} = {self.created_at},
                    {"owner": <10} = {self.user},
                    {"team": <10} = \"{self.team}\",
                    {"cluster_id": <10} = {self.cluster_id},
                )"""
            )
        else:
            res = textwrap.dedent(
                f"""\
                {self.__class__.__name__}(
                    {"name": <10} = \"{self.name}\",
                    {"version": <10} = {self.version},
                    {"source": <10} = \"{self.source}\",
                    {"compression": <10} = {self.compression},
                    {"owner": <10} = {self.user},
                    {"team": <10} = \"{self.team}\",
                    {"cluster_id": <10} = {self.cluster_id},
                )"""
            )
        return res

    @affirm(is_not_shallow, is_not_deleted)
    def __str__(self):
        return repr(self)

    @affirm(is_not_shallow, is_not_deleted)
    def __eq__(self, other: 'Datastore'):
        # TODO - handling team's datastore equality here is probably not the best. We should
        #  delegate that to the backend when Project lands
        # need to handle case where attributes of a DataStore are not `User` or `Team`
        # classes. This is the case before the datastore is uploaded.
        self_team = self._team.team_id if hasattr(self._team, 'team_id') else self._team
        other_team = other._team.team_id if hasattr(other._team, 'team_id') else self._team

        self_owner = self._user.user_id if hasattr(self._user, 'user_id') else self._user
        other_owner = other._user.user_id if hasattr(other._user, 'user_id') else other.user

        return (
            self.__class__.__qualname__ == other.__class__.__qualname__ and self._name == other._name
            and self._version == other._version and self_owner == other_owner and self_team == other_team
        )

    @affirm(is_not_shallow, is_not_deleted)
    def __hash__(self):
        return hash((
            self._name, self._id, self._version, self._size, self._created_at, self._snapshot_status, self._user,
            self._team, self._source, self._compression, self._cluster_id, self._is_deleted, self._is_created
        ))

    # ---------------------  Internal Methods ----------------------

    @affirm(is_not_shallow, is_not_deleted)
    def _update_status(self):
        """Refreshes the``snapshot_status`` attribute by querying the Grid API.
        """
        tid = None if self._team is None else self.team.team_id
        all_dstore_data = get_available_datastores(team_id=tid)

        for dstore_data in all_dstore_data:
            if dstore_data['id'] == self._id:
                self._snapshot_status = dstore_data['snapshotStatus'] or "unknown"

    # -------------------  Public Facing Methods ----------------------

    @affirm(is_not_shallow, is_not_deleted, is_created)
    def delete(self):
        """Deletes the datastore from the grid system.
        """
        delete_datastore(name=self.name, version=self.version, cluster=self.cluster_id)
        self._is_deleted = True

    @affirm(is_not_shallow, is_not_created)
    def upload(self):
        """Uploads the contents of the directories referenced by this datastore instance to Grid.

        Depending on your internet connection this may be a potentially long running process.
        If uploading is inturupsed, the upload session can be resumed by initializing this
        ``Datastore`` object again with the same parameters repeating the call to ``upload()``.
        """
        session = create_datastore_session(
            name=self.name,
            source=self.source,
            compression=self.compression,
            cluster=self.cluster_id,
        )
        session.upload()

        # Since we can't query just a single datastore by it's ID (THIS IS INSANE!)
        # we just grab the entire world of datastores and check to see which one
        # has the same attributes as this (as determined by this classes __eq__ method)

        time.sleep(1)  # give the backend time to process the new record.
        for datastore in list_datastores(is_global=True):
            if datastore == self:
                self._name = datastore._name
                self._id = datastore._id
                self._version = datastore._version
                self._source = datastore._source
                self._compression = datastore._compression
                self._snapshot_status = datastore._snapshot_status
                self._created_at = datastore._created_at
                self._user = datastore._user
                self._team = datastore._team
                self._cluster_id = datastore._cluster_id
                self._size = datastore._size

                self._is_deleted = False
                self._is_created = True


def list_datastores(is_global: bool = False) -> List[Datastore]:
    """List datastores for user / teams

    Parameters
    ----------
    is_global:
        if True, returns a list of datastores of the everyone in the team
    """
    datastores = []
    user_dstores_data = get_available_datastores()
    for dstore_data in user_dstores_data:
        user = User(
            username=dstore_data['userDetails']['username'],
            user_id=dstore_data['userDetails']['id'],
            first_name=dstore_data['userDetails']['firstName'],
            last_name=dstore_data['userDetails']['lastName'],
        )
        dstore = Datastore._from_existing(  # noqa
            name=dstore_data['name'],
            version=int(dstore_data['version']),
            owner=user,
            team=None,
            size=dstore_data['size'],
            created_at=datetime.fromisoformat(dstore_data['createdAt']),
            snapshot_status=dstore_data['snapshotStatus'],
            datastore_id=dstore_data['id'],
            cluster_id=dstore_data['clusterId'],
        )
        datastores.append(dstore)

    # If ``include_teams`` is set, add datastores registered to the team.

    teams = []
    if is_global:
        teams_data = get_user_teams()
        for team_data in teams_data:
            members = {}
            for member_data in team_data['members']:
                user = User(
                    username=member_data['username'],
                    user_id=member_data['id'],
                    first_name=member_data['firstName'],
                    last_name=member_data['lastName'],
                )
                members[user.username] = user

            team = Team(
                team_id=team_data['id'],
                name=team_data['name'],
                created_at=team_data['createdAt'],
                role=team_data['role'],
                members=members,
            )
            teams.append(team)

    for team in teams:
        team_dstores_data = get_available_datastores(team_id=team.team_id)
        for dstore_data in team_dstores_data:
            user = User(
                user_id=dstore_data['userDetails']['id'],
                username=dstore_data['userDetails']['username'],
                first_name=dstore_data['userDetails']['firstName'],
                last_name=dstore_data['userDetails']['lastName']
            )

            dstore = Datastore._from_existing(
                name=dstore_data['name'],
                version=int(dstore_data['version']),
                owner=user,
                team=team,
                size=dstore_data['size'],
                created_at=datetime.fromisoformat(dstore_data['createdAt']),
                snapshot_status=dstore_data['snapshotStatus'],
                datastore_id=dstore_data['id'],
                cluster_id=dstore_data['clusterId'],
            )
            datastores.append(dstore)
    return datastores


def parse_name_from_source(source) -> str:
    """Parses datastore name from source if name isn't provided"""
    try:
        parse_result = urlparse(source)
    except ValueError:
        raise click.ClickException("Invalid source for datastore, please input only a local filepath or valid url")

    path = Path(parse_result.path)
    base = path.name.split(".")[0]
    return base.lower().strip()



from grid.sdk import (
    Datastore,
    Experiment,
    get_instance_types,
    list_datastores,
    list_sessions,
    login,
    Run,
    list_runs,
    Actions,
    ScratchSpace,
    Resources,
    Session,
)

__all__ = [
    "login", "list_datastores", "Run", "Experiment", "list_runs", "Actions", "ScratchSpace", "Resources", "Datastore",
    "get_instance_types", "list_sessions", "Session"
]



import json
import os
from pathlib import Path
from typing import Optional

from grid.sdk._gql.queries import get_grid_user_id
from grid.sdk.auth import Credentials
from grid.sdk import env
from grid.sdk.api import list_clusters
from grid.sdk.client.grid_gql import gql_client


def login(
    username: Optional[str] = None,
    user_id: Optional[str] = None,
    api_key: Optional[str] = None,
    *,
    _url=env.GRID_URL  # noqa - For testing purposes only. not user facing.
):
    """
    Log in with your grid.ai credentials for usage of the SDK in the running process.

    All parameters are optional. Calling ``login()`` without parameters will check if
    the ``GRID_USER_ID`` and ``GRID_API_KEY`` env vars have been set (using those if
    available), otherwise it will check for the file ``credentials.json`` in the
    machines ``$HOME/.grid`` directory (if it exists).

    If no credentials have been stored, then you must pass in your API key and either
    your username or user id (if you know it). Your user id and API key can be found
    by navigating to *** INSERT INSTRUCTIONS TO FIND KEY HERE ***.

    Parameters
    ----------
    username
        your grid username. This is either be your github username or email address,
        depending on what you use when signing into the grid platform at:
        https://platform.grid.ai

    user_id
        Your grid user id. This can be found by ** INSERT INSTRUCTIONS HERE **
    api_key
        Your grid API key. This can be found by ** INSERT INSTRUCTIONS HERE **
    """
    if user_id and api_key:
        creds = Credentials(user_id=user_id, api_key=api_key)
        _create_credentials_file(creds)
    elif username and api_key:
        creds = Credentials(user_id=username, api_key=api_key)
        client = gql_client(url=_url + "/graphql", creds=creds)
        user_id = get_grid_user_id(client=client)
        creds = Credentials(user_id=user_id, api_key=api_key)
        _create_credentials_file(creds)
    elif (username or user_id) or api_key:
        raise ValueError("Either (`user_id` OR `username`) AND `api_key` need to be set, or none set.")
    else:
        creds = Credentials.from_locale()

    # adding context to the user settings
    # this will overwrite the context user set - when we have the set context command.
    # solution to that problem is to have DEFAULT_CONTEXT and SELECTED_CONTEXT
    settings_path = Path(env.GRID_SETTINGS_PATH)
    if not settings_path.exists():
        env.write_default_settings(settings_path)
    user_settings = json.load(settings_path.open())
    cluster_list = list_clusters(is_global=True)
    default_cluster = cluster_list.default_cluster
    user_settings['context'] = default_cluster
    user_settings['grid_url'] = _url
    with settings_path.open('w') as file:
        json.dump(user_settings, file, ensure_ascii=False, indent=4)

    # resetting the env variables
    os.environ['GRID_USER_ID'] = creds.user_id
    os.environ['GRID_API_KEY'] = creds.api_key
    os.environ['GRID_URL'] = _url
    os.environ['GRID_CLUSTER_ID'] = default_cluster
    env.reset_global_variables()

    creds = Credentials.from_locale()
    # warmup cache. can take ~2 sec on first call
    _ = gql_client(url=_url + '/graphql', creds=creds)
    return True


def _create_credentials_file(creds: Credentials):
    Path(env.GRID_CREDENTIAL_PATH).parent.mkdir(parents=True, exist_ok=True)
    with Path(env.GRID_CREDENTIAL_PATH).open('w') as file:
        json.dump({'UserID': creds.user_id, 'APIKey': creds.api_key}, file, ensure_ascii=False, indent=4)



from dataclasses import dataclass
from datetime import datetime, timedelta
import json
import os
from pathlib import Path
from typing import Union
import uuid

import grpc
import jwt

from grid.sdk import env


@dataclass(unsafe_hash=True)
class Credentials:
    user_id: str
    api_key: str

    @classmethod
    def from_locale(cls) -> "Credentials":
        """Instantiates the credentials using implicit locale.

        First use environment variables, otherwise look for credentials stored in file.

        Returns
        -------
        Credentials
            instantiated credentials object.
        """
        # if user has environment variables, use that
        user_id = os.getenv('GRID_USER_ID')
        api_key = os.getenv('GRID_API_KEY')
        grid_url = os.getenv('GRID_URL')
        if grid_url:
            env.GRID_URL = grid_url
        if user_id and api_key:
            return cls(user_id=user_id, api_key=api_key)

        # otherwise overwrite look for credentials stored locally as a file
        if os.getenv("CI"):
            p = Path.home() / ".grid" / "credentials.json"
        else:
            p = Path(os.getenv('GRID_CREDENTIAL_PATH', Path.home() / ".grid" / "credentials.json"))

        if not p.exists():
            raise PermissionError('No credentials available. Did you login?')
        with p.open() as f:
            credentials = json.load(f)

        return cls(
            user_id=credentials['UserID'],
            api_key=credentials['APIKey'],
        )

    @classmethod
    def from_file(cls, path: Union[str, Path]) -> "Credentials":
        """Instantiates the credentials using specified file path.

        Parameters
        ----------
        path
            file path to the config yaml or json on disk.

        Returns
        -------
        Credentials
            instantiated credentials object.
        """
        p = Path(path).absolute()
        if not p.exists():
            raise PermissionError('No credentials available. Did you login?')
        with p.open() as f:
            credentials = json.load(f)

        return cls(
            user_id=credentials['UserID'],
            api_key=credentials['APIKey'],
        )


class GrpcAuth(grpc.AuthMetadataPlugin):
    def __init__(self, credentials: "Credentials"):
        self._creds = credentials

    @property
    def _bearer_token(self) -> str:
        if os.getenv("GRID_AUTH_TOKEN"):
            return os.getenv("GRID_AUTH_TOKEN")

        dt_now = datetime.utcnow()
        # TODO(rusenask): Delete this when you migrate everything to nice JWTs
        token = jwt.encode(
            payload={
                "aud": ["grid"],
                "exp": dt_now + timedelta(seconds=30),
                "iat": dt_now,
                "iss": "grid-cli",
                "jti": str(uuid.uuid4()),
                "nbf": dt_now,
                "sub": self._creds.user_id
            },
            key=self._creds.api_key,
            algorithm="HS256"
        )

        return token.decode("ascii")

    def __call__(self, context: "grpc.AuthMetadataContext", callback: "grpc.AuthMetadataPluginCallback") -> None:
        callback((('authorization', f"Bearer {self._bearer_token}"), ), None)



import textwrap
from datetime import datetime
from typing import Dict, NamedTuple, Optional, List

from grid.sdk.datastores import Datastore
from grid.sdk.run_components import Actions, Resources, ScratchSpace
from grid.sdk import env
from grid.sdk.rest.client import GridRestClient
from grid.sdk.affirmations import affirm, is_not_deleted, is_created, is_not_created, is_not_shallow
from grid.sdk.rest.experiments import get_experiment_from_id, get_experiment_by_name, update_experiment, \
    delete_experiment
from grid.sdk.user import User
from grid.sdk.client import create_swagger_client

from grid.openapi.models import (V1ExperimentState, V1PackageManager, Externalv1Experiment)
from grid.sdk.utilities import MAGIC_NAME


class ExperimentName(NamedTuple):
    name: str
    username: Optional[str]

    @classmethod
    def from_external_format(cls, name: str):
        """User can pass experiments as username:experiment_name to fetch other users experiments.
        """
        username = None
        split = name.split(":")
        if len(split) > 2:
            raise ValueError(f"Error while parsing {name}. Use the format <username>:<experiment-name>")
        elif len(split) == 2:
            username = split[0]
            name = split[1]

        return cls(name=name, username=username)


class Experiment:
    _client: GridRestClient

    _created_at: datetime
    _description: str
    _id: str
    _name: str
    _updated_at: datetime
    _actions: Actions
    _cluster_id: str
    _run_command: str
    _datastore: Datastore
    _datastore_mount_dir: str
    _desired_state: str
    _dry_run: bool
    _env: Dict[str, str]
    _runner_framework: str
    _runner_dockerfile: str
    _runner_requirements_file: str
    _runner_package_manager: str
    _instance_type: str
    _resources: Resources
    _run_controller_id: str
    _scratch: ScratchSpace
    _source_code: str
    _use_spot: bool
    _user_id: str
    _cost: float
    _deleted_timestamp: datetime
    _finish_timestamp: datetime
    _hourly_cost: float
    _message: str
    _status: str
    _start_timestamp: datetime
    _user: User

    _is_deleted: bool
    _is_created: bool
    _is_shallow: bool

    def __init__(
        self,
        name: str,
        cluster_id: Optional[str] = None,
    ):
        # --------------------------------------------------------------------------- #
        #    This should be the first block that goes into the constructor of the     #
        #    resource object. This is to ensure that the resource object can be       #
        #    created from another internal call without filling the values and        #
        #    let __getattr__ picks up the instruction for refreshing values from ID   #
        self._client = GridRestClient(api_client=create_swagger_client())
        cluster_id = cluster_id or env.CONTEXT
        self._is_shallow = False
        self._cluster_id = cluster_id  # will be overwritten if comes from the config
        if name == MAGIC_NAME:
            self._is_shallow = True
            self._is_created = False
            self._is_deleted = False
            return
        #                                                                             #
        # --------------------------------------------------------------------------- #

        try:
            exp = get_experiment_by_name(client=self._client, cluster_id=self._cluster_id, experiment_name=self._name)
            self._setup_from_response(exp)
            return
        except KeyError:
            raise RuntimeError(
                f"Experiment {name} does not exist. If you are here "
                f"for creating experiment, use Run(...) since experiments "
                f"are grouped under Run object and can only be created "
                f"using the Run abstraction"
            ) from None

    @property
    @affirm(is_not_shallow)
    def exists(self) -> bool:
        return self._is_created

    @classmethod
    def _from_id(cls, exp_id: str, cluster_id: Optional[str] = env.CONTEXT) -> "Experiment":
        instance = cls(name=MAGIC_NAME, cluster_id=cluster_id)
        instance._id = exp_id
        return instance

    def _setup_from_response(self, exp: 'Externalv1Experiment'):
        self._is_deleted = exp.status.phase == V1ExperimentState.DELETED
        self._is_created = True
        self._is_shallow = False

        spec = exp.spec
        status = exp.status
        image = spec.image

        # TODO - setup datastore
        self._name = exp.name
        self._id = exp.id
        self._created_at = exp.created_at
        self._description = exp.description
        self._updated_at = exp.updated_at

        # spec attributes
        self._cluster_id = spec.cluster_id
        self._command = spec.command
        self._desired_state = str(spec.desired_state)
        self._dry_run = spec.dry_run
        self._env = spec.env
        self._instance_type = spec.instance_type
        self._run_controller_id = spec.run_controller_id
        self._source_code = spec.source_code
        self._use_spot = spec.use_spot
        self._user_id = spec.user_id
        # TODO - set actions, resources and scratch

        # image attributes
        self._runner_framework = image.framework
        self._runner_dockerfile = image.dockerfile
        # TODO - understand why these values could be empty from the BE
        self._runner_requirements_file = image.dependency_file_info.path if image.dependency_file_info else None
        self._runner_package_manager = image.dependency_file_info.package_manager if image.dependency_file_info else None

        # status attributes
        self._cost = status.cost
        self._deleted_timestamp = status.deleted_timestamp
        self._finish_timestamp = status.finish_timestamp
        self._hourly_cost = status.hourly_cost
        self._message = status.message
        self._status = str(status.phase)
        self._start_timestamp = status.start_timestamp

        if not hasattr(self, '_user') or self._user is None:
            # TODO - fetch user from ID - we don't have that in BE
            self._user = User(user_id=spec.user_id, username="", first_name="", last_name="")

    @property
    def id(self) -> str:
        return self._id

    @affirm(is_not_shallow, is_not_deleted)
    def __repr__(self):
        return textwrap.dedent(
            f"""\
            {self.__class__.__name__}(
                {"name": <18} = {self._name},
                {"desired_state": <18} = {self._desired_state},
                {"status": <18} = {self._status},
                {"source_code": <18} = {self._source_code},
                {"command": <18} = {self._command},
                {"user": <18} = {self.user.user_id},
                {"cluster_id": <18} = {self._cluster_id},
                {"created_at": <18} = {self._created_at},
                {"start_timestamp": <18} = {self._start_timestamp},
                {"finished_timestamp": <18} = {self._finish_timestamp},
            )"""
        )

    def _unshallow(self):
        if not self._is_shallow:
            raise RuntimeError('Experiment is already unshallow')
        if not hasattr(self, '_id') or self._id is None:
            raise RuntimeError("Cannot unshallow resource without a valid Experiment id")
        self._is_shallow = False
        exp = get_experiment_from_id(self._client, experiment_id=self._id, cluster_id=self._cluster_id)
        self._setup_from_response(exp)

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def created_at(self) -> datetime:
        """created_at of the experiment configured as part of the run.
        """
        return self._created_at

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def description(self) -> str:
        """description of the experiment configured as part of the run.
        """
        return self._description

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def name(self) -> str:
        """name of the experiment configured as part of the run.
        """
        return self._name

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def actions(self) -> Actions:
        """actions of the experiment configured as part of the run.
        """
        return self._actions

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def cluster_id(self) -> str:
        """cluster_id of the experiment configured as part of the run.
        """
        return self._cluster_id

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def command(self) -> List[str]:
        """command of the experiment configured as part of the run.
        """
        return self._command

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def datastore(self) -> Datastore:
        """datastores of the experiment configured as part of the run.
        """
        return self._datastore

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def datastore_mount_dir(self) -> str:
        return self._datastore_mount_dir

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def desired_state(self) -> str:
        """desired_state of the experiment configured as part of the run.
        """
        return self._desired_state

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def dry_run(self) -> bool:
        """dry_run of the experiment configured as part of the run.
        """
        return self._dry_run

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def env(self) -> Dict[str, str]:
        """env of the experiment configured as part of the run.
        """
        return self._env

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def runner_framework(self) -> str:  # from _image: 'Externalv1ImageSpec'
        """runner_framework of the experiment configured as part of the run.
        """
        return self._runner_framework

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def runner_dockerfile(self) -> str:  # from _image: 'Externalv1ImageSpec'
        """runner_dockerfile of the experiment configured as part of the run.
        """
        return self._runner_dockerfile

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def runner_requirements_file(self) -> str:  # from _image: 'Externalv1ImageSpec'
        """runner_requirements_file of the experiment configured as part of the run.
        """
        return self._runner_requirements_file

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def runner_package_manager(self) -> 'V1PackageManager':  # from _image: 'Externalv1ImageSpec'
        """runner_package_manager of the experiment configured as part of the run.
        """
        return self._runner_package_manager

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def instance_type(self) -> str:
        """instance_type of the experiment configured as part of the run.
        """
        return self._instance_type

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def resources(self) -> Resources:
        """resources of the experiment configured as part of the run.
        """
        return self._resources

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def run_controller_id(self) -> str:
        """run_controller_id of the experiment configured as part of the run.
        """
        return self._run_controller_id

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def scratch(self) -> ScratchSpace:
        """scratch of the experiment configured as part of the run.
        """
        return self._scratch

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def source_code(self) -> str:
        """source_code of the experiment configured as part of the run.
        """
        return self._source_code

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def use_spot(self) -> bool:
        """use_spot of the experiment configured as part of the run.
        """
        return self._use_spot

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def user(self) -> User:
        """user who owns the of the experiment/run.
        """
        return self._user

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def cost(self) -> float:
        """cost of the experiment configured as part of the run.
        """
        return self._cost

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def deleted_timestamp(self) -> datetime:
        """deleted_timestamp of the experiment configured as part of the run.
        """
        return self._deleted_timestamp

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def finish_timestamp(self) -> datetime:
        """finish_timestamp of the experiment configured as part of the run.
        """
        return self._finish_timestamp

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def hourly_cost(self) -> float:
        """hourly_cost of the experiment configured as part of the run.
        """
        return self._hourly_cost

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def message(self) -> str:
        """message of the experiment configured as part of the run.
        """
        return self._message

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def status(self) -> str:
        """status of the experiment configured as part of the run.
        """
        return self._status

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def start_timestamp(self) -> datetime:
        """start_timestamp of the experiment configured as part of the run.
        """
        return self._start_timestamp

    @affirm(is_not_shallow, is_not_deleted, is_created)
    def cancel(self) -> bool:
        """Request the grid platform cancels the experiment
        """
        self._is_deleted = True
        # do not perform POST if we know that we don't have to
        exp = update_experiment(
            client=self._client,
            cluster_id=self._cluster_id,
            experiment_id=self._id,
            desired_state=V1ExperimentState.CANCELLED
        )
        self._setup_from_response(exp)
        return True

    @affirm(is_not_shallow, is_not_deleted, is_created)
    def delete(self) -> bool:
        """Request the grid platform cancels the experiment
        """
        # do not perform POST if we know that we don't have to
        delete_experiment(client=self._client, cluster_id=self._cluster_id, experiment_id=self._id)
        self._is_deleted = True
        self._is_created = False
        return True



from typing import Dict, Optional, List

from grid.openapi import Externalv1Datastore, Externalv1DatastoreSpec, V1GetDatastoreResponse, V1ListDatastoresResponse
from grid.sdk.rest.client import GridRestClient
from grid.sdk.rest.exceptions import throw_with_message


@throw_with_message
def datastore_id_from_name(c: GridRestClient, cluster_id: str, name: str, version: Optional[str] = None) -> str:
    """Find the id of a datastore with some name (and optionally version) on a cluster.

    Parameters
    ----------
    c
        Client
    cluster_id
        which cluster should be used to find the datastore in.
    name
        the name of the datastore to find the ID of
    version
        The version of the datastore with ``name`` to find the ID of.
        NOTE: If no ``version`` argument is present, then the maximum
        version of the datastore will be used.

    Returns
    -------
    str
       The ID of the datastore.
    """
    dstores: V1ListDatastoresResponse = c.datastore_service_list_datastores(cluster_id=cluster_id, available=True)

    datastore_versions: Dict[str, Externalv1Datastore] = {}

    for dstore in dstores.datastores:
        dstore: Externalv1Datastore
        if dstore.name == name:
            spec: Externalv1DatastoreSpec = dstore.spec
            datastore_versions[spec.version] = dstore

    if version is None:
        # use the max version available
        version = max(datastore_versions.keys())

    try:
        return datastore_versions[version].id
    except KeyError:
        raise KeyError(f"no datastore exists with name: {name}")


@throw_with_message
def get_datastore_from_id(c: GridRestClient, datastore_id: str, cluster_id: str) -> V1GetDatastoreResponse:
    dstore: V1GetDatastoreResponse = c.datastore_service_get_datastore(cluster_id=cluster_id, id=datastore_id)
    return dstore


@throw_with_message
def get_datastore_from_name(client: GridRestClient, cluster_id: str, datastore_name: str, version: int):
    datastores = list_datastores(client=client, cluster_id=cluster_id)
    ds_version_to_name_map = {}
    for datastore in datastores:
        if datastore.name == datastore_name:
            # return if the version is the same
            if version and datastore.spec.version == version:
                return datastore
            ds_version_to_name_map[datastore.spec.version] = datastore.id
    # if version is not specified, return the latest version
    if not version and len(ds_version_to_name_map) > 0:
        max_version = max(ds_version_to_name_map.keys())
        return ds_version_to_name_map[max_version]
    raise KeyError(f'Datastore with name {datastore_name} and version {version} not found')


@throw_with_message
def list_datastores(client: GridRestClient, cluster_id: str, user_ids=None) -> List[Externalv1Datastore]:
    user_ids = user_ids or []
    resp = client.datastore_service_list_datastores(cluster_id=cluster_id, user_ids=user_ids, available=True)
    return resp.datastores


def datastore_dsn_from_id(id: str) -> str:
    """Return the DSN of the datastore

    Parameters
    ----------
    id
        datastore ID to convert into DSN.

    Returns
    -------
    str
        DSN form of the datastore ID.
    """
    return f"datastore://grid/{id}"


def datastore_id_from_dsn(dsn: str) -> str:
    """Return the id of a datastore from a DSN.

    Parameters
    ----------
    dsn
        DSN string to convert into an ID

    Returns
    -------
    str
        ID of the datastore DSN string.
    """
    # convert ``datastore://grid/{id}`` -> ``['datastore:', '', 'grid', '{id}']``
    # convert ``datastore://grid/{id}/`` -> ``['datastore:', '', 'grid', '{id}', '']``
    # ... (we want the last element)
    parts = dsn.split('/')
    if dsn.endswith('/'):
        if len(parts) < 2:
            raise RuntimeError(f"Internal Error. invalid datastore dsn format while parsing ID. dsn={dsn}")
        return parts[-2]
    return parts[-1]



# coding: utf-8

"""
    external/v1/external_session_service.proto

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: version not set
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    NOTE
    ----
    standard swagger-codegen-cli for this python client has been modified
    by custom templates. The purpose of these templates is to include
    typing information in the API and Model code. Please refer to the
    main grid repository for more info
"""


from __future__ import absolute_import

import re  # noqa: F401
from typing import TYPE_CHECKING, Any

# python 2 and python 3 compatibility library
import six

from grid.openapi.api_client import ApiClient

if TYPE_CHECKING:
    from datetime import datetime
    from grid.openapi.models import *


class DatastoreServiceApi(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    Ref: https://github.com/swagger-api/swagger-codegen
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient()
        self.api_client = api_client

    def datastore_service_complete_datastore_presigned_urls_upload(self, cluster_id: 'str', datastore_id: 'str', upload_id: 'str', body: 'Body3', **kwargs) -> 'V1CompleteDatastorePresignedUrlsUploadResponse':  # noqa: E501
        """datastore_service_complete_datastore_presigned_urls_upload  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_complete_datastore_presigned_urls_upload(cluster_id, datastore_id, upload_id, body, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str datastore_id: (required)
        :param str upload_id: (required)
        :param Body3 body: (required)
        :return: V1CompleteDatastorePresignedUrlsUploadResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.datastore_service_complete_datastore_presigned_urls_upload_with_http_info(cluster_id, datastore_id, upload_id, body, **kwargs)  # noqa: E501
        else:
            (data) = self.datastore_service_complete_datastore_presigned_urls_upload_with_http_info(cluster_id, datastore_id, upload_id, body, **kwargs)  # noqa: E501
            return data

    def datastore_service_complete_datastore_presigned_urls_upload_with_http_info(self, cluster_id: 'str', datastore_id: 'str', upload_id: 'str', body: 'Body3', **kwargs) -> 'V1CompleteDatastorePresignedUrlsUploadResponse':  # noqa: E501
        """datastore_service_complete_datastore_presigned_urls_upload  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_complete_datastore_presigned_urls_upload_with_http_info(cluster_id, datastore_id, upload_id, body, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str datastore_id: (required)
        :param str upload_id: (required)
        :param Body3 body: (required)
        :return: V1CompleteDatastorePresignedUrlsUploadResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['cluster_id', 'datastore_id', 'upload_id', 'body']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method datastore_service_complete_datastore_presigned_urls_upload" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'cluster_id' is set
        if self.api_client.client_side_validation and ('cluster_id' not in params or
                                                       params['cluster_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `cluster_id` when calling `datastore_service_complete_datastore_presigned_urls_upload`")  # noqa: E501
        # verify the required parameter 'datastore_id' is set
        if self.api_client.client_side_validation and ('datastore_id' not in params or
                                                       params['datastore_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `datastore_id` when calling `datastore_service_complete_datastore_presigned_urls_upload`")  # noqa: E501
        # verify the required parameter 'upload_id' is set
        if self.api_client.client_side_validation and ('upload_id' not in params or
                                                       params['upload_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `upload_id` when calling `datastore_service_complete_datastore_presigned_urls_upload`")  # noqa: E501
        # verify the required parameter 'body' is set
        if self.api_client.client_side_validation and ('body' not in params or
                                                       params['body'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `body` when calling `datastore_service_complete_datastore_presigned_urls_upload`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'cluster_id' in params:
            path_params['clusterId'] = params['cluster_id']  # noqa: E501
        if 'datastore_id' in params:
            path_params['datastoreId'] = params['datastore_id']  # noqa: E501
        if 'upload_id' in params:
            path_params['uploadId'] = params['upload_id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        if 'body' in params:
            body_params = params['body']
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/{clusterId}/datastores/{datastoreId}/presigned-urls/{uploadId}', 'POST',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1CompleteDatastorePresignedUrlsUploadResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def datastore_service_create_datastore(self, spec_cluster_id: 'str', body: 'V1CreateDatastoreRequest', **kwargs) -> 'V1CreateDatastoreResponse':  # noqa: E501
        """datastore_service_create_datastore  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_create_datastore(spec_cluster_id, body, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str spec_cluster_id: (required)
        :param V1CreateDatastoreRequest body: (required)
        :return: V1CreateDatastoreResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.datastore_service_create_datastore_with_http_info(spec_cluster_id, body, **kwargs)  # noqa: E501
        else:
            (data) = self.datastore_service_create_datastore_with_http_info(spec_cluster_id, body, **kwargs)  # noqa: E501
            return data

    def datastore_service_create_datastore_with_http_info(self, spec_cluster_id: 'str', body: 'V1CreateDatastoreRequest', **kwargs) -> 'V1CreateDatastoreResponse':  # noqa: E501
        """datastore_service_create_datastore  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_create_datastore_with_http_info(spec_cluster_id, body, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str spec_cluster_id: (required)
        :param V1CreateDatastoreRequest body: (required)
        :return: V1CreateDatastoreResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['spec_cluster_id', 'body']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method datastore_service_create_datastore" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'spec_cluster_id' is set
        if self.api_client.client_side_validation and ('spec_cluster_id' not in params or
                                                       params['spec_cluster_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `spec_cluster_id` when calling `datastore_service_create_datastore`")  # noqa: E501
        # verify the required parameter 'body' is set
        if self.api_client.client_side_validation and ('body' not in params or
                                                       params['body'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `body` when calling `datastore_service_create_datastore`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'spec_cluster_id' in params:
            path_params['spec.clusterId'] = params['spec_cluster_id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        if 'body' in params:
            body_params = params['body']
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/{spec.clusterId}/datastores', 'POST',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1CreateDatastoreResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def datastore_service_create_datastore_presigned_urls(self, cluster_id: 'str', datastore_id: 'str', body: 'Body2', **kwargs) -> 'V1CreateDatastorePresignedUrlsResponse':  # noqa: E501
        """datastore_service_create_datastore_presigned_urls  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_create_datastore_presigned_urls(cluster_id, datastore_id, body, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str datastore_id: (required)
        :param Body2 body: (required)
        :return: V1CreateDatastorePresignedUrlsResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.datastore_service_create_datastore_presigned_urls_with_http_info(cluster_id, datastore_id, body, **kwargs)  # noqa: E501
        else:
            (data) = self.datastore_service_create_datastore_presigned_urls_with_http_info(cluster_id, datastore_id, body, **kwargs)  # noqa: E501
            return data

    def datastore_service_create_datastore_presigned_urls_with_http_info(self, cluster_id: 'str', datastore_id: 'str', body: 'Body2', **kwargs) -> 'V1CreateDatastorePresignedUrlsResponse':  # noqa: E501
        """datastore_service_create_datastore_presigned_urls  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_create_datastore_presigned_urls_with_http_info(cluster_id, datastore_id, body, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str datastore_id: (required)
        :param Body2 body: (required)
        :return: V1CreateDatastorePresignedUrlsResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['cluster_id', 'datastore_id', 'body']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method datastore_service_create_datastore_presigned_urls" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'cluster_id' is set
        if self.api_client.client_side_validation and ('cluster_id' not in params or
                                                       params['cluster_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `cluster_id` when calling `datastore_service_create_datastore_presigned_urls`")  # noqa: E501
        # verify the required parameter 'datastore_id' is set
        if self.api_client.client_side_validation and ('datastore_id' not in params or
                                                       params['datastore_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `datastore_id` when calling `datastore_service_create_datastore_presigned_urls`")  # noqa: E501
        # verify the required parameter 'body' is set
        if self.api_client.client_side_validation and ('body' not in params or
                                                       params['body'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `body` when calling `datastore_service_create_datastore_presigned_urls`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'cluster_id' in params:
            path_params['clusterId'] = params['cluster_id']  # noqa: E501
        if 'datastore_id' in params:
            path_params['datastoreId'] = params['datastore_id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        if 'body' in params:
            body_params = params['body']
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/{clusterId}/datastores/{datastoreId}/presigned-urls', 'POST',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1CreateDatastorePresignedUrlsResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def datastore_service_delete_datastore(self, cluster_id: 'str', id: 'str', **kwargs) -> 'V1DeleteDatastoreResponse':  # noqa: E501
        """datastore_service_delete_datastore  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_delete_datastore(cluster_id, id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str id: (required)
        :return: V1DeleteDatastoreResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.datastore_service_delete_datastore_with_http_info(cluster_id, id, **kwargs)  # noqa: E501
        else:
            (data) = self.datastore_service_delete_datastore_with_http_info(cluster_id, id, **kwargs)  # noqa: E501
            return data

    def datastore_service_delete_datastore_with_http_info(self, cluster_id: 'str', id: 'str', **kwargs) -> 'V1DeleteDatastoreResponse':  # noqa: E501
        """datastore_service_delete_datastore  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_delete_datastore_with_http_info(cluster_id, id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str id: (required)
        :return: V1DeleteDatastoreResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['cluster_id', 'id']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method datastore_service_delete_datastore" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'cluster_id' is set
        if self.api_client.client_side_validation and ('cluster_id' not in params or
                                                       params['cluster_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `cluster_id` when calling `datastore_service_delete_datastore`")  # noqa: E501
        # verify the required parameter 'id' is set
        if self.api_client.client_side_validation and ('id' not in params or
                                                       params['id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `id` when calling `datastore_service_delete_datastore`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'cluster_id' in params:
            path_params['clusterId'] = params['cluster_id']  # noqa: E501
        if 'id' in params:
            path_params['id'] = params['id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/{clusterId}/datastores/{id}', 'DELETE',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1DeleteDatastoreResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def datastore_service_get_datastore(self, cluster_id: 'str', id: 'str', **kwargs) -> 'V1GetDatastoreResponse':  # noqa: E501
        """datastore_service_get_datastore  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_get_datastore(cluster_id, id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str id: (required)
        :return: V1GetDatastoreResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.datastore_service_get_datastore_with_http_info(cluster_id, id, **kwargs)  # noqa: E501
        else:
            (data) = self.datastore_service_get_datastore_with_http_info(cluster_id, id, **kwargs)  # noqa: E501
            return data

    def datastore_service_get_datastore_with_http_info(self, cluster_id: 'str', id: 'str', **kwargs) -> 'V1GetDatastoreResponse':  # noqa: E501
        """datastore_service_get_datastore  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_get_datastore_with_http_info(cluster_id, id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str id: (required)
        :return: V1GetDatastoreResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['cluster_id', 'id']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method datastore_service_get_datastore" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'cluster_id' is set
        if self.api_client.client_side_validation and ('cluster_id' not in params or
                                                       params['cluster_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `cluster_id` when calling `datastore_service_get_datastore`")  # noqa: E501
        # verify the required parameter 'id' is set
        if self.api_client.client_side_validation and ('id' not in params or
                                                       params['id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `id` when calling `datastore_service_get_datastore`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'cluster_id' in params:
            path_params['clusterId'] = params['cluster_id']  # noqa: E501
        if 'id' in params:
            path_params['id'] = params['id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/{clusterId}/datastores/{id}', 'GET',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1GetDatastoreResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def datastore_service_list_datastore_events(self, cluster_id: 'str', id: 'str', **kwargs) -> 'V1ListDatastoreEventsResponse':  # noqa: E501
        """datastore_service_list_datastore_events  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_list_datastore_events(cluster_id, id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str id: (required)
        :return: V1ListDatastoreEventsResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.datastore_service_list_datastore_events_with_http_info(cluster_id, id, **kwargs)  # noqa: E501
        else:
            (data) = self.datastore_service_list_datastore_events_with_http_info(cluster_id, id, **kwargs)  # noqa: E501
            return data

    def datastore_service_list_datastore_events_with_http_info(self, cluster_id: 'str', id: 'str', **kwargs) -> 'V1ListDatastoreEventsResponse':  # noqa: E501
        """datastore_service_list_datastore_events  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_list_datastore_events_with_http_info(cluster_id, id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str id: (required)
        :return: V1ListDatastoreEventsResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['cluster_id', 'id']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method datastore_service_list_datastore_events" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'cluster_id' is set
        if self.api_client.client_side_validation and ('cluster_id' not in params or
                                                       params['cluster_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `cluster_id` when calling `datastore_service_list_datastore_events`")  # noqa: E501
        # verify the required parameter 'id' is set
        if self.api_client.client_side_validation and ('id' not in params or
                                                       params['id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `id` when calling `datastore_service_list_datastore_events`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'cluster_id' in params:
            path_params['clusterId'] = params['cluster_id']  # noqa: E501
        if 'id' in params:
            path_params['id'] = params['id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/{clusterId}/datastores/{id}/events', 'GET',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1ListDatastoreEventsResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def datastore_service_list_datastores(self, cluster_id: 'str', **kwargs) -> 'V1ListDatastoresResponse':  # noqa: E501
        """datastore_service_list_datastores  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_list_datastores(cluster_id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str page_token:
        :param str limit:
        :param list[str] user_ids:
        :param bool available:
        :return: V1ListDatastoresResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.datastore_service_list_datastores_with_http_info(cluster_id, **kwargs)  # noqa: E501
        else:
            (data) = self.datastore_service_list_datastores_with_http_info(cluster_id, **kwargs)  # noqa: E501
            return data

    def datastore_service_list_datastores_with_http_info(self, cluster_id: 'str', **kwargs) -> 'V1ListDatastoresResponse':  # noqa: E501
        """datastore_service_list_datastores  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_list_datastores_with_http_info(cluster_id, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str cluster_id: (required)
        :param str page_token:
        :param str limit:
        :param list[str] user_ids:
        :param bool available:
        :return: V1ListDatastoresResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['cluster_id', 'page_token', 'limit', 'user_ids', 'available']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method datastore_service_list_datastores" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'cluster_id' is set
        if self.api_client.client_side_validation and ('cluster_id' not in params or
                                                       params['cluster_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `cluster_id` when calling `datastore_service_list_datastores`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'cluster_id' in params:
            path_params['clusterId'] = params['cluster_id']  # noqa: E501

        query_params = []
        if 'page_token' in params:
            query_params.append(('pageToken', params['page_token']))  # noqa: E501
        if 'limit' in params:
            query_params.append(('limit', params['limit']))  # noqa: E501
        if 'user_ids' in params:
            query_params.append(('userIds', params['user_ids']))  # noqa: E501
            collection_formats['userIds'] = 'multi'  # noqa: E501
        if 'available' in params:
            query_params.append(('available', params['available']))  # noqa: E501

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/{clusterId}/datastores', 'GET',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1ListDatastoresResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)

    def datastore_service_update_datastore(self, spec_cluster_id: 'str', id: 'str', body: 'Body6', **kwargs) -> 'V1UpdateDatastoreResponse':  # noqa: E501
        """datastore_service_update_datastore  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_update_datastore(spec_cluster_id, id, body, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str spec_cluster_id: (required)
        :param str id: (required)
        :param Body6 body: (required)
        :return: V1UpdateDatastoreResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """
        kwargs['_return_http_data_only'] = True
        if kwargs.get('async_req'):
            return self.datastore_service_update_datastore_with_http_info(spec_cluster_id, id, body, **kwargs)  # noqa: E501
        else:
            (data) = self.datastore_service_update_datastore_with_http_info(spec_cluster_id, id, body, **kwargs)  # noqa: E501
            return data

    def datastore_service_update_datastore_with_http_info(self, spec_cluster_id: 'str', id: 'str', body: 'Body6', **kwargs) -> 'V1UpdateDatastoreResponse':  # noqa: E501
        """datastore_service_update_datastore  # noqa: E501

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.datastore_service_update_datastore_with_http_info(spec_cluster_id, id, body, async_req=True)
        >>> result = thread.get()

        :param async_req bool
        :param str spec_cluster_id: (required)
        :param str id: (required)
        :param Body6 body: (required)
        :return: V1UpdateDatastoreResponse
                 If the method is called asynchronously,
                 returns the request thread.
        """

        all_params = ['spec_cluster_id', 'id', 'body']  # noqa: E501
        all_params.append('async_req')
        all_params.append('_return_http_data_only')
        all_params.append('_preload_content')
        all_params.append('_request_timeout')

        params = locals()
        for key, val in six.iteritems(params['kwargs']):
            if key not in all_params:
                raise TypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method datastore_service_update_datastore" % key
                )
            params[key] = val
        del params['kwargs']
        # verify the required parameter 'spec_cluster_id' is set
        if self.api_client.client_side_validation and ('spec_cluster_id' not in params or
                                                       params['spec_cluster_id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `spec_cluster_id` when calling `datastore_service_update_datastore`")  # noqa: E501
        # verify the required parameter 'id' is set
        if self.api_client.client_side_validation and ('id' not in params or
                                                       params['id'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `id` when calling `datastore_service_update_datastore`")  # noqa: E501
        # verify the required parameter 'body' is set
        if self.api_client.client_side_validation and ('body' not in params or
                                                       params['body'] is None):  # noqa: E501
            raise ValueError("Missing the required parameter `body` when calling `datastore_service_update_datastore`")  # noqa: E501

        collection_formats = {}

        path_params = {}
        if 'spec_cluster_id' in params:
            path_params['spec.clusterId'] = params['spec_cluster_id']  # noqa: E501
        if 'id' in params:
            path_params['id'] = params['id']  # noqa: E501

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        body_params = None
        if 'body' in params:
            body_params = params['body']
        # HTTP header `Accept`
        header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # HTTP header `Content-Type`
        header_params['Content-Type'] = self.api_client.select_header_content_type(  # noqa: E501
            ['application/json'])  # noqa: E501

        # Authentication setting
        auth_settings = []  # noqa: E501

        return self.api_client.call_api(
            '/v1/core/{spec.clusterId}/datastores/{id}', 'PUT',
            path_params,
            query_params,
            header_params,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type='V1UpdateDatastoreResponse',  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get('async_req'),
            _return_http_data_only=params.get('_return_http_data_only'),
            _preload_content=params.get('_preload_content', True),
            _request_timeout=params.get('_request_timeout'),
            collection_formats=collection_formats)
